/**********************************************************************
 *
 * cgiwrapper.cpp -- output pages using the cgi protocol
 * Copyright (C) 1999  The New Zealand Digital Library Project
 *
 * A component of the Greenstone digital library software
 * from the New Zealand Digital Library Project at the
 * University of Waikato, New Zealand.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *********************************************************************/

#include "gsdlconf.h"
#include "cgiwrapper.h"
#include "fileutil.h"
#include "recptconfig.h"

#if defined(GSDL_USE_OBJECTSPACE)
#  include <ospace/std/iostream>
#  include <ospace/std/fstream>
#elif defined(GSDL_USE_IOS_H)
#  include <iostream.h>
#  include <fstream.h>
#else
#  include <iostream>
#  include <fstream>
#endif

static void format_error_string (text_t &errorpage, const text_t &errortext, bool debug) {

  errorpage.clear();

  if (debug) {
    errorpage += "\n";
    errorpage += "ERROR: " + errortext;
    errorpage += "\n";
    
  } else {

    errorpage += "Content-type: text/html\n\n";
    
    errorpage += "<html>\n";
    errorpage += "<head>\n";
    errorpage += "<title>Error</title>\n";
    errorpage += "</head>\n";
    errorpage += "<body>\n";
    errorpage += "<h2>Oops!</h2>\n";
    errorpage += errortext;
    errorpage += "</body>\n";
    errorpage += "</html>\n";
  }
}

static void page_errorsitecfg (text_t &errorpage, bool debug, int mode) {

  text_t errortext;

  if (mode == 0) {
    errortext += "The gsdlsite.cfg configuration file could not be found. This\n";
    errortext += "file should contain configuration information relating to this\n";
    errortext += "site's setup.\n";

  } else if (mode == 1) {
    errortext += "The gsdlsite.cfg configuration file does not contain a valid\n";
    errortext += "gsdlhome entry.\n";
  }

  if (debug) {
    errortext += "gsdlsite.cfg should reside in the directory from which the\n";
    errortext += "library executable was run.\n";
  } else {
    errortext += "gsdlsite.cfg should reside in the same directory as the library\n";
    errortext += "executable file.\n";
  }

  format_error_string (errorpage, errortext, debug);
}


static void page_errormaincfg (const text_t &gsdlhome, bool debug, text_t &errorpage) {

  text_t errortext;
  text_t main_cfg_file = filename_cat (gsdlhome, "etc", "main.cfg");
  errortext += "The main.cfg configuration file could not be found. This file\n";
  errortext += "should contain configuration information relating to the\n";
  errortext += "setup of the interface. As this receptionist is not being run\n";
  errortext += "in collection specific mode the file should reside at\n";
  errortext += main_cfg_file + ".\n";
  format_error_string (errorpage, errortext, debug);
}


static void page_errorinit (const text_t &gsdlhome, bool debug, text_t &errorpage) {

  text_t errortext = "An error occurred during the initialisation of the Greenstone Digital\n";
  errortext += "Library software. It is likely that the software has not been setup\n";
  errortext += "correctly.\n";

  text_t error_file = filename_cat (gsdlhome, "etc", "error.txt");
  char *efile = error_file.getcstr();
  ifstream errin (efile);
  delete efile;
  if (errin) {
    errortext += "The error log, " + error_file + ", contains the\n";
    errortext += "following information:\n\n";
    if (!debug) errortext += "<pre>\n";

    char c;
    errin.get(c);
    while (!errin.eof ()) {
      errortext.push_back(c);
      errin.get(c);
    }
    
    if (!debug) errortext += "</pre>\n";

    errin.close();

  } else {
    errortext += "Please consult " + error_file + " for more information.\n";
  }

  format_error_string (errorpage, errortext, debug);
}

static void page_errorparseargs (const text_t &gsdlhome, bool debug, text_t &errorpage) {

  text_t errortext = "An error occurred during the parsing of the cgi arguments.\n";

  text_t error_file = filename_cat (gsdlhome, "etc", "error.txt");
  char *efile = error_file.getcstr();
  ifstream errin (efile);
  delete efile;
  if (errin) {
    errortext += "The error log, " + error_file + ", contains the\n";
    errortext += "following information:\n\n";
    if (!debug) errortext += "<pre>\n";

    char c;
    errin.get(c);
    while (!errin.eof ()) {
      errortext.push_back(c);
      errin.get(c);
    }
    if (!debug) errortext += "</pre>\n";
    errin.close();

  } else {
    errortext += "Please consult " + error_file + " for more information.\n";
  }

  format_error_string (errorpage, errortext, debug);
}

static void page_errorcgipage (const text_t &gsdlhome, bool debug, text_t &errorpage) {

  text_t errortext = "An error occurred during the construction of the cgi page.\n";

  text_t error_file = filename_cat (gsdlhome, "etc", "error.txt");
  char *efile = error_file.getcstr();
  ifstream errin (efile);
  delete efile;
  if (errin) {
    errortext += "The error log, " + error_file + ", contains the\n";
    errortext += "following information:\n\n";
    if (!debug) errortext += "<pre>\n";

    char c;
    errin.get(c);
    while (!errin.eof ()) {
      errortext.push_back(c);
      errin.get(c);
    }
    if (!debug) errortext += "</pre>\n";
    errin.close();

  } else {
    errortext += "Please consult " + error_file + " for more information.\n";
  }

  format_error_string (errorpage, errortext, debug);
}

static void print_debug_info (receptionist &recpt) {

  outconvertclass text_t2ascii;
  const recptconf &configinfo = recpt.get_configinfo ();
  text_t etc_dir = filename_cat (configinfo.gsdlhome, "etc");

  cout << "\n";
  cout << text_t2ascii
       << "------------------------------------------------------------\n"
       << "Configuration and initialization completed successfully.\n"
       << "  Note that more debug information may be available in the\n"
       << "  initialization and error log error.txt in " << etc_dir << ".\n"
       << "------------------------------------------------------------\n\n";
  
  cout << text_t2ascii << "gsdlhome=" << configinfo.gsdlhome << "\n";
  if (!configinfo.gdbmhome.empty())
    cout << text_t2ascii << "gdbmhome=" << configinfo.gdbmhome << "\n";
  cout << text_t2ascii << "httpprefix=" << configinfo.httpprefix << "\n";
  cout << text_t2ascii << "httpimg=" << configinfo.httpimg << "\n";
  cout << text_t2ascii << "gwcgi=" << configinfo.gwcgi << "\n"
       << "  Note that unless gwcgi has been set from a configuration\n"
       << "  file it is dependent on environment variables set by your\n"
       << "  webserver. Therefore it may not have the same value when run\n"
       << "  from the command line as it would be when run from your\n"
       << "  web server.\n";
  if (configinfo.usecookies) 
    cout << "cookies are enabled\n";
  else 
    cout << "cookies are disabled\n";
  if (configinfo.logcgiargs) 
    cout << "logging is enabled\n";
  else 
    cout << "logging is disabled\n";
  cout << "------------------------------------------------------------\n\n";

  text_tset::const_iterator this_mfile = configinfo.macrofiles.begin();
  text_tset::const_iterator end_mfile = configinfo.macrofiles.end();
  cout << "Macro Files:\n"
       << "------------\n";
  text_t mfile;
  bool found;
  while (this_mfile != end_mfile) {
    cout << text_t2ascii << *this_mfile;
    int spaces = (22 - (*this_mfile).size());
    if (spaces < 2) spaces = 2;
    text_t outspaces;
    for (int i = 0; i < spaces; i++) outspaces.push_back (' ');
    cout << text_t2ascii << outspaces;

    mfile = filename_cat (configinfo.gsdlhome, "macros", *this_mfile);
    if (file_exists (mfile)) {
      cout << text_t2ascii << "found (" << mfile << ")\n";
    } else {
      cout << text_t2ascii << "NOT FOUND\n";
    }

    this_mfile ++;
  }

  cout << "\n------------------------------------------------------------\n";
  cout << "------------------------------------------------------------\n\n";
  cout << "receptionist running in command line debug mode\n";
  cout << "enter cgi arguments as name=value pairs (e.g. 'a=p&p=home'):\n";

}

void cgiwrapper (receptionist &recpt) {
  
  bool debug = false;
  const recptconf &configinfo = recpt.get_configinfo ();

  // get the query string
  text_t argstr = "";
  cgiargsclass args;
  char *aURIStr;
  char *request_method_str = getenv("REQUEST_METHOD");
  char *content_length_str = getenv("CONTENT_LENGTH");
  if (request_method_str != NULL && strcmp(request_method_str, "POST") == 0 &&
      content_length_str != NULL)  {
    // POST form data
    int content_length = text_t(content_length_str).getint();
    if (content_length > 0) {
      char c;
      do {
	cin.get(c);
	if (cin.eof()) break;
	argstr.push_back (c);
	content_length--;
      } while (content_length > 0);
    }
      
  } else {
    aURIStr = getenv("QUERY_STRING");
    if ((request_method_str != NULL && strcmp(request_method_str, "GET") == 0)
	|| aURIStr != NULL) {
      // GET form data
      if (aURIStr != NULL) argstr = aURIStr;
    } else {
      // debugging from command line
      debug = true;
    }
  }

  if (debug) {
    cout << "Configuring Greenstone...\n";
    cout << flush;
  }

  text_t errorpage;
  outconvertclass text_t2ascii;

  // set defaults
  recpt.configure ("httpimg", "/gsdl/images");
  char *script_name = getenv("SCRIPT_NAME");
  if (script_name != NULL) recpt.configure("gwcgi", script_name);
  else recpt.configure("gwcgi", "/gsdl");

  // read in the configuration files.
  text_t gsdlhome;
  if (!site_cfg_read (recpt, gsdlhome)) {
    // couldn't find the site configuration file
    page_errorsitecfg (errorpage, debug, 0);
  } else if (gsdlhome.empty()) {
    // no gsdlhome in gsdlsite.cfg
    page_errorsitecfg (errorpage, debug, 1);
  } else if (!directory_exists(gsdlhome)) {
    // gsdlhome not a valid directory
    page_errorsitecfg (errorpage, debug, 1);
  } else if (!main_cfg_read (recpt, gsdlhome)) {
    // couldn't find the main configuration file
    page_errormaincfg (gsdlhome, debug, errorpage);
  }

  if (errorpage.empty()) {

    // initialise the library software
    if (debug) {
      cout << "Initializing...\n";
      cout << flush;
    }

    text_t error_file = filename_cat (gsdlhome, "etc", "error.txt");
    char *eout = error_file.getcstr();
    ofstream errout (eout, ios::app);
    delete eout;
    if (!recpt.init(errout)) {
      // an error occurred during the initialisation
      errout.close();
      page_errorinit(gsdlhome, debug, errorpage);
    }
    errout.close();
  }

  if (debug && errorpage.empty()) {
    // get query string from command line
    print_debug_info (recpt);
    char cinURIStr[1024];
    cin.get(cinURIStr, 1024);
    argstr = cinURIStr;
  }

#define pageout cout

  // if there has been no error so far, perform the production of the
  // output page
  if (errorpage.empty()) {
    text_t error_file = filename_cat (gsdlhome, "etc", "error.txt");
    char *eout = error_file.getcstr();
    ofstream errout (eout, ios::app);
    delete eout;

#if defined(__WIN32__) && defined(GSDL_USE_IOS_H)
    // old Windows compilers (VC++4.2)
    cerr = errout;
#else
    // can't do this anymore according to c++ standard...
    // cerr = errout;
    // ... but can do this instead
    streambuf* errbuf = cerr.rdbuf(errout.rdbuf());
#endif

    // parse the cgi arguments and produce the resulting page if there
    // has been no errors so far
    if (!recpt.parse_cgi_args (argstr, args, errout)) {
      errout.close ();
      page_errorparseargs(gsdlhome, debug, errorpage);
    } else {
      // produce the output page

      if (!recpt.produce_cgi_page (args, pageout, errout)) {
	errout.close ();
	page_errorcgipage(gsdlhome, debug, errorpage);
      }
      recpt.log_cgi_args (args, errout);
      errout.close ();
    }

#if !defined(__WIN32__) || !defined(GSDL_USE_IOS_H)
    // restore the cerr buffer
    cerr.rdbuf(errbuf);
#endif
  }

  // there was an error, output the error page
  if (!errorpage.empty()) {
    pageout << text_t2ascii << errorpage;
    errorpage.clear();
  }
  pageout << flush;

  return;
}
